import sys
from os import path

import nonebot
from nonebot.adapters.cqhttp import Bot
from nonebot.log import default_format, logger


nonebot.init()
driver = nonebot.get_driver()
driver.register_adapter('cqhttp', Bot)
app = nonebot.get_asgi()

nonebot.load_builtin_plugins()
nonebot.load_plugin('nonebot_plugin_apscheduler')
# 删除 haruka_bot 导入，否则 nonebot 导入时会忽略
del sys.modules['haruka_bot']
nonebot.load_plugin("haruka_bot")

# Modify some config / config depends on loaded configs
# 
# config = nonebot.get_driver().config
# do something...

logger.add(path.join('log', "error.log"),
           rotation="00:00",
           retention='1 week',
           diagnose=False,
           level="ERROR",
           format=default_format,
           encoding='utf-8'
           )

def run():
    nonebot.run(app="haruka_bot.cli.bot:app")



from nonebot import on_command, on_notice
from nonebot.adapters.cqhttp import Bot, Event, GroupDecreaseNoticeEvent
from nonebot.permission import SUPERUSER
from nonebot.adapters.cqhttp.permission import GROUP_ADMIN, GROUP_OWNER

from .bilireq import BiliReq
from .config import Config
from .utils import permission_check, to_me
from .version import __version__


add_uid = on_command('添加主播', rule=to_me() & permission_check, priority=5)

@add_uid.handle()
async def get_args(bot: Bot, event: Event, state: dict):
    args = str(event.message).strip()
    if args:
        state['uid'] = args

@add_uid.got('uid', prompt='请输入主播的 uid')
async def _(bot: Bot, event: Event, state: dict):
    uid = state['uid']
    with Config(event) as config:
        await add_uid.finish(await config.add_uid(uid))


delete_uid = on_command('删除主播', rule=to_me() & permission_check, priority=5)

@delete_uid.handle()
async def get_args(bot: Bot, event: Event, state: dict):
    args = str(event.message).strip()
    if args:
        state['uid'] = args

@delete_uid.got('uid', prompt='请输入主播的 uid')
async def _(bot: Bot, event: Event, state: dict):
    uid = state['uid']
    with Config(event) as config:
        await delete_uid.finish(await config.delete_uid(uid))


uid_list = on_command('主播列表', rule=to_me() & permission_check, priority=5)

@uid_list.handle()
async def _(bot: Bot, event: Event, state: dict):
    with Config(event) as config:
        await uid_list.finish(await config.uid_list())


dynamic_on = on_command('开启动态', rule=to_me() & permission_check, priority=5)

@dynamic_on.handle()
async def handle_first_recive(bot: Bot, event: Event, state: dict):
    args = str(event.message).strip()
    if args:
        state['uid'] = args

@dynamic_on.got('uid', prompt='请输入主播的 uid')
async def _(bot: Bot, event: Event, state: dict):
    uid = state['uid']
    with Config(event) as config:
        msg = await config.set('dynamic', uid, True)
    await dynamic_on.finish(msg.replace('name', '开启动态'))


dynamic_off = on_command('关闭动态', rule=to_me() & permission_check, priority=5)

@dynamic_off.handle()
async def handle_first_recive(bot: Bot, event: Event, state: dict):
    args = str(event.message).strip()
    if args:
        state['uid'] = args

@dynamic_off.got('uid', prompt='请输入主播的 uid')
async def _(bot: Bot, event: Event, state: dict):
    uid = state['uid']
    with Config(event) as config:
        msg = await config.set('dynamic', uid, False)
    await dynamic_off.finish(msg.replace('name', '关闭动态'))


live_on = on_command('开启直播', rule=to_me() & permission_check, priority=5)

@live_on.handle()
async def handle_first_recive(bot: Bot, event: Event, state: dict):
    args = str(event.message).strip()
    if args:
        state['uid'] = args

@live_on.got('uid', prompt='请输入主播的 uid')
async def _(bot: Bot, event: Event, state: dict):
    uid = state['uid']
    with Config(event) as config:
        msg = await config.set('live', uid, True)
    await live_on.finish(msg.replace('name', '开启直播'))
    

live_off = on_command('关闭直播', rule=to_me() & permission_check, priority=5)

@live_off.handle()
async def handle_first_recive(bot: Bot, event: Event, state: dict):
    args = str(event.message).strip()
    if args:
        state['uid'] = args

@live_off.got('uid', prompt='请输入主播的 uid')
async def _(bot: Bot, event: Event, state: dict):
    uid = state['uid']
    with Config(event) as config:
        msg = await config.set('live', uid, False)
    await live_off.finish(msg.replace('name', '关闭直播'))


at_on = on_command('开启全体', rule=to_me(), 
    permission=GROUP_OWNER | GROUP_ADMIN | SUPERUSER, 
    priority=5)

@at_on.handle()
async def handle_first_recive(bot: Bot, event: Event, state: dict):
    args = str(event.message).strip()
    if args:
        state['uid'] = args

@at_on.got('uid', prompt='请输入主播的 uid')
async def _(bot: Bot, event: Event, state: dict):
    uid = state['uid']
    with Config(event) as config:
        msg = await config.set('at', uid, True)
    await at_on.finish(msg.replace('name', '开启全体'))


at_off = on_command('关闭全体', rule=to_me(), 
    permission=GROUP_OWNER | GROUP_ADMIN | SUPERUSER, 
    priority=5)

@at_off.handle()
async def handle_first_recive(bot: Bot, event: Event, state: dict):
    args = str(event.message).strip()
    if args:
        state['uid'] = args

@at_off.got('uid', prompt='请输入主播的 uid')
async def _(bot: Bot, event: Event, state: dict):
    uid = state['uid']
    with Config(event) as config:
        msg = await config.set('at', uid, False)
    await at_off.finish(msg.replace('name', '关闭全体'))


permission_on = on_command('开启权限', rule=to_me(), 
    permission=GROUP_OWNER | GROUP_ADMIN | SUPERUSER, 
    priority=5)

@permission_on.handle()
async def _(bot: Bot, event: Event, state: dict):
    with Config(event) as config:
        msg = await config.set_permission(True)
    await permission_on.finish(msg.replace('name', '开启权限'))


func_list = [
    '主播列表', '开启权限', '关闭权限', '添加主播', '删除主播', '开启动态',
    '关闭动态', '开启直播', '关闭直播', '开启全体', '关闭全体', '版本信息']

permission_off = on_command('关闭权限', rule=to_me(), 
    permission=GROUP_OWNER | GROUP_ADMIN | SUPERUSER, 
    priority=5)

@permission_off.handle()
async def _(bot: Bot, event: Event, state: dict):
    with Config(event) as config:
        msg = await config.set_permission(False)
    await permission_on.finish(msg.replace('name', '关闭权限'))

get_version = on_command('版本信息', rule=to_me(), priority=5)

@get_version.handle()
async def _(bot: Bot, event: Event, state: dict):
    message = f"当前 HarukaBot 版本：{__version__}\n" +\
        "\n使用中遇到问题欢迎加群反馈，\n" +\
        "群号：629574472\n" +\
        "\n常见问题：https://www.haruka-bot.live/usage/faq.html"
    await get_version.finish(message)

no_permission = on_command(func_list[0],
    aliases=set(func_list[1:]), rule=to_me(), priority=20)

@no_permission.handle()
async def _(bot: Bot, event: Event, state: dict):
    await permission_on.finish("权限不足，目前只有管理员才能使用")


help = on_command('帮助', rule=to_me(), priority=5)

@help.handle()
async def _(bot: Bot, event: Event, state: dict):
    message = "DD机目前支持的功能有：\n\n"
    for name in func_list:
        message += name
        if not name.endswith(('列表', '权限', '版本信息')):
            message += " uid"
        message += '\n'
    message += "\n命令中的uid需要替换为对应主播的uid，注意是uid不是直播间id\n" + \
        "\n群聊默认开启权限，只有管理员或机器人主人才能触发指令\n" + \
        "\n所有群聊/私聊的推送都是分开的，在哪里添加就会在哪里推送"
    await help.finish(message)


group_decrease = on_notice(priority=5)

@group_decrease.handle()
async def _(bot: Bot, event: GroupDecreaseNoticeEvent, state: dict):
    if event.self_id == event.user_id:
        event.message_type = 'group'
        c = Config(event)
        await c.delete_push_list()


# login = on_command('测试登录', rule=to_me(), permission=SUPERUSER, 
#     priority=5)

# @login.handle()
# async def _(bot: Bot, event: Event, state: dict):
#     b = BiliReq()
#     await login.send(f"[CQ:image,file=base64://{await b.get_qr()}]")
#     await login.send(str(await b.qr_login()))
    


import json
from datetime import datetime

import nonebot
from nonebot.adapters.cqhttp import MessageEvent
from tinydb import TinyDB, Query

from .utils import get_path
from .version import __version__
from packaging.version import Version


class Config():
    """操作 config.json 文件"""

    def __init__(self, event:MessageEvent=None):
        self._init(event)
    
    def __enter__(self, event:MessageEvent=None):
        self._init(event)
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.config.close()
    
    def _init(self, event=None):
        self.config = TinyDB(get_path('config.json'), encoding='utf-8')
        self.uids = self.config.table('uids')
        self.groups = self.config.table('groups')
        self.uid_lists = self.config.table('uid_lists')
        self.login = self.config.table('login')
        self.version = self.config.table('version')

        if event:
            self.bot_id = str(event.self_id)
            self.type = event.message_type
            self.type_id = str(event.group_id) if self.type == 'group' else str(event.user_id)
    
    def uid_exist(self, uid, type_id=False):
        q = Query()
        if type_id:
            r = self.config.get((q.uid == uid) & (q.type == self.type) & (q.type_id == self.type_id))
        else:
            r = self.config.get((q.uid == uid))
        return r
    
    def add_admin(self):
        if self.type == 'private':
            return
        q = Query()
        if not self.groups.contains(q.group_id == self.type_id):
            self.groups.insert({'group_id': self.type_id, 'admin': True})

    def update_uid_lists(self):
        if 'uid_lists' not in self.config.tables():
            self.uid_lists.insert_multiple([
                {
                    'dynamic': [],
                    'index': 0
                },
                {
                    'live': [],
                    'index': 0
                }
            ])
        q = Query()
        r = self.config.search(q.dynamic == True)
        dynamic = list(set([c['uid'] for c in r]))
        self.uid_lists.update({'dynamic': dynamic}, q.dynamic.exists())
        r = self.config.search(q.live == True)
        live = list(set([c['uid'] for c in r]))
        self.uid_lists.update({'live': live}, q.live.exists())

    async def add_uid(self, uid):
        """添加主播"""
        
        r = self.uid_exist(uid, True)
        if r: # 已经存在
            return f"请勿重复添加 {r['name']}（{r['uid']}）"
        
        self.add_admin()
        r = self.uid_exist(uid)
        if r: # 当前账号没订阅，但是其他账号添加过这个 uid
            name = r['name']
        else: # 没有账号订阅过这个 uid
            from .bilireq import BiliReq
            br = BiliReq()
            try: # 检测 uid 是否有效（逻辑需要修改）
                user_info = await br.get_info(uid)
                name = user_info["name"]
            except:
                return "请输入有效的uid"
            
        self.config.insert({
            'uid': uid,
            'name': name,
            'type': self.type, 
            'type_id': self.type_id,
            'live': True,
            'dynamic': True,
            'at': False,
            'bot_id': self.bot_id
            })
        self.update_uid_lists()
        return f"已添加 {name}（{uid}）"
    
    async def delete_uid(self, uid):
        """删除主播"""

        r = self.uid_exist(uid, True)
        if not r:
            return "删除失败，uid 不存在"
        q = Query()
        self.config.remove(
            (q.uid == uid) & 
            (q.type == self.type) & 
            (q.type_id == self.type_id))
        self.update_uid_lists()
        return f"已删除 {r['name']}（{uid}）"
    
    async def delete_push_list(self):
        """删除指定对象的推送列表"""

        q = Query()
        self.config.remove(
            (q.type == self.type) &
            (q.type_id == self.type_id))
        self.update_uid_lists()

    async def uid_list(self):
        """主播列表"""

        q = Query()
        r = self.config.search((q.type == self.type) & (q.type_id == self.type_id))
        message = "以下为当前的订阅列表：\n\n"
        for c in r:
            message += (
                f"【{c['name']}】" +
                f"直播推送：{'开' if c['live'] else '关'}，" +
                f"动态推送：{'开' if c['dynamic'] else '关'}" +
                f"（{c['uid']}）\n"
            )
        return message

    async def set(self, func, uid, status):
        """开关各项功能"""

        if func == 'at' and self.type == 'private':
            return "只有群里才能name"

        r = self.uid_exist(uid, True)
        if not r:
            return "name失败，uid 不存在"
        
        if r[func] == status:
            return "请勿重复name"

        q = Query()
        self.config.update({func: status}, (q.uid == uid) & (q.type == self.type) & (q.type_id == self.type_id))
        self.update_uid_lists()
        return f"已name，{r['name']}（{r['uid']}）"

    async def set_permission(self, status):
        """设置权限"""

        if self.type == 'private':
            return "只有群里才能name"
        
        q = Query()
        r = self.groups.get(q.group_id == self.type_id)
        if (not r and status) or (r and r['admin'] == status):
            return "请勿重复name"
        if not self.groups.contains(q.group_id == self.type_id):
            self.groups.insert({'group_id': self.type_id, 'admin': status})
        else:
            self.groups.update({'admin': status}, q.group_id == self.type_id)
        return "已name，只有管理员才能使用" if status else "已name，所有人都能使用"

    def next_uid(self, func):
        """获取下一位爬取的 uid"""
        
        q = Query()
        r = self.uid_lists.get(q[func].exists())
        if not r: # 一次都没有添加过，uid_list 还没有创建
            return None
        index = r['index']
        uid_list = r[func]

        if not uid_list: # uid_list 为空
            return None
        
        if index >= len(uid_list):
            uid = uid_list[0]
            index = 1
        else:
            uid = uid_list[index]
            index += 1
        self.uid_lists.update({'index': index}, q[func].exists())
        return uid
    
    
    def get_uid_list(self, name):
        """获取需要爬取的 UID 列表"""

        q = Query()
        r = self.uid_lists.get(q[name].exists())
        if not r:
            return []
        return r[name]


    def get_push_list(self, uid, func):
        """获取推送列表"""

        q = Query()
        return self.config.search((q.uid == uid) & (q[func] == True))

    def get_admin(self, group_id):
        q = Query()
        if not self.groups.contains(q.group_id == group_id):
            return True
        return self.groups.get(q.group_id == group_id)['admin']

    @classmethod
    def get_name(cls, uid):
        """获取 uid 对应的昵称"""

        q = Query()
        return (cls().config.get(q.uid == str(uid)))['name']
    
    def read(self):
        """读取用户注册信息"""

        with open(get_path('config.json'), encoding='utf-8-sig') as f:
            text = f.read()
        self.json = json.loads(text)
        return self.json

    def backup(self):
        """备份当前配置文件"""

        # FIXME 如果 config.json 不存在，不备份
        self.read()
        backup_name = f"config.{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.json.bak"
        with open(get_path(backup_name), 'w', encoding='utf-8') as f:
            f.write(json.dumps(self.json, ensure_ascii=False, indent=4))
        return True
    
    @classmethod
    def get_login(cls):
        """获取登录信息"""

        if 'login' not in cls().config.tables():
            tokens = {
                'access_token': '',
                'refresh_token': ''
            }
            cls().login.insert(tokens)
        else:
            tokens = cls().login.all()[0]
        if tokens == {'access_token': '', 'refresh_token': ''}:
            return None
        return tokens

    @classmethod
    def update_login(cls, tokens):
        """更新登录信息"""

        cls().login.update({
            'access_token': tokens['access_token'],
            'refresh_token': tokens['refresh_token']
        })

    def new_version(self):
        if 'version' not in self.config.tables():
            self.version.insert({'version': __version__})
            return True
        current_version = Version(__version__)
        old_version = Version(self.version.all()[0]['version'])
        return current_version > old_version
    
    def update_version(self):
        self.version.update({'version': __version__})
                

    @classmethod
    async def update_config(cls):
        """升级为 TinyDB"""

        with Config() as config:
            if 'status' in config.config.tables():
                config.backup()
                config.config.drop_tables()
                
                for c_type, config_type in {'group': 'groups', 'private': 'users'}.items():
                    config.type = c_type
                    for type_id, type_config in config.json[config_type].items():
                        config.type_id = type_id
                        uids = type_config['uid']
                        for uid, sets in uids.items():
                            config.bot_id = config.json['uid'][uid][config_type][config.type_id]
                            await config.add_uid(uid)
                            for func, status in sets.items():
                                await config.set(func, uid, status)
                        if 'admin' in type_config and not type_config['admin']:
                            await config.set_permission(False)
            if config.new_version():
                config.backup()
                config.update_version()


nonebot.get_driver().on_startup(Config.update_config)



import nonebot

try:
    nonebot.get_driver()
    from . import utils
    from . import config_manager
    from . import live_pusher
    from . import dynamic_pusher
    from . import auto_agree
except ValueError:
    pass

from .version import __version__, VERSION
